import pandas as pd
from sklearn import datasets
from tidyframe import gather, spread

iris = datasets.load_iris()
df = pd.DataFrame(iris['data'], columns=iris.feature_names)
df['target'] = iris.target
df['target2'] = list(map(lambda x: 1 if x<1 else 0, df.target))
col_gather = df.columns[:4]
df_short = df.head()
df_short_gather = gather(df_short[col_gather])
df_short_gather2 = gather(df_short, col_gather )

def test_spread_basic():
    spread(df_short_gather, ['index'], 'key')

def test_spread_string():
    spread(df_short_gather, 'index', 'key')

def test_spread_mulit_value():
    df_short_gather2 = df_short_gather
    df_short_gather2['value2'] = df_short_gather2['value'] + 1
    spread(df_short_gather, 'index', 'key')

def test_spread_mulit_key():
    df_short_gather['key_with_index'] = df_short_gather['key'] + "_" + df_short_gather['index'].astype(str)
    spread(df_short_gather, 'index', ['key', 'key_with_index'])

def test_spread_mulit_row_key():
    df_short_gather['index2'] = df_short_gather['index'] + 1 
    spread(df_short_gather, ['index', 'index2'], 'key')



""" Convert Pandas DataFrame to nest DataFrame """

import copy as cp
from functools import partial
import pandas as pd
import numpy as np

def nest(df, columns=[], columns_minus=[], columns_between=[], key='data', copy=False):
    """
    Nest repeated values

    Parameters
    ----------
    df: DataFrameGroupBy or DataFrame
    columns: list or index, nest columns
    columns_minus: list or index, columns which do not want to nest
                   (must choose one of columns and columns_minus)
    columns_between: list with length 2, assigin nest columns between to two columns
    copy: False, return DataFrame using copy.deepcopy
    """
    assert isinstance(df, (pd.core.frame.DataFrame, pd.core.groupby.DataFrameGroupBy)), "Must be DataFrame"
    if len(columns) > 0:
        assert len(columns_minus) == 0 and len(columns_between) == 0, "Using Parameter columns then columns_minus and column_between must not use"
    if len(columns_minus) > 0:
        assert len(columns) == 0 and len(columns_between) == 0, "Using Parameter columns_minus then columns and columns_between must not use"
    if len(columns_between) > 0:
        assert len(columns_between) == 2, "lenth of columns_between must be 2"
        assert len(columns) == 0 and len(columns_minus) == 0, "Using Parameter columns_between then columns_minus and between must not use"

    if isinstance(df, pd.core.frame.DataFrame):
        if len(columns) > 0:
            if isinstance(columns, pd.core.indexes.base.Index):
                columns_nest = columns.tolist()
            else:
                columns_nest = columns
            columns_group = df.columns.difference(columns_nest).tolist()
            df_g = df.groupby(columns_group)
            data = [[index, group[columns_nest]] for index, group in df_g]
        elif len(columns_minus) > 0:
            if isinstance(columns_minus, pd.core.indexes.base.Index):
                columns_group = columns_minus.tolist()
            else:
                columns_group = columns_minus
            columns_nest = df.columns.difference(columns_group).tolist()
            df_g = df.groupby(columns_group)
            data = [[index, group[columns_nest]] for index, group in df_g]
        else:
            index_start = np.where(df.columns == columns_between[0])[0][0]
            index_end = np.where(df.columns == columns_between[1])[0][0]
            assert index_start < index_end, "columns_between order error"
            columns_nest = df.columns[index_start:(index_end+1)].tolist()
            columns_group = df.columns.difference(columns_nest).tolist()
            df_g = df.groupby(columns_group)
            data = [[index, group[columns_nest]] for index, group in df_g]
    else:
        columns_group = list(df.dtypes.index.names)
        columns_nest = list(df.dtypes.columns)
        data = [[index, group[columns_nest]] for index, group in df]
    outer = list(map(lambda x: x[0], data))
    df_return = pd.DataFrame(outer, columns=columns_group)
    df_return[key] = list(map(lambda x: x[1][columns_nest], data))
    if copy:
        return cp.deepcopy(df_return)
    else:
        return df_return

def unnest(df, drop=[], copy=False):
    """
    Inverse Nest DataFrame

    Parameters
    ----------
    df: DataFrame with Series of Dataframe
    drop: list of column which do not return
    """
    df_check = df.applymap(lambda x: isinstance(x, pd.DataFrame))
    columns_nest = df_check.columns[df_check.sum() == df_check.shape[0]].tolist()
    if len(columns_nest)>0:
        if len(columns_nest) == 1:
            repeat_times = list(map(lambda x: x.shape[0], df[columns_nest[0]]))
            columns_group = df_check.columns.difference(columns_nest)
            df_return = pd.DataFrame(df[columns_group].as_matrix().repeat(repeat_times, axis=0), columns=columns_group)
            df_return = pd.concat([df_return, pd.concat([*df[columns_nest[0]].tolist()])], axis=1)
            if copy:
                return cp.deepcopy(df_return[df_return.columns.difference(drop)])
            else:
                return df_return[df_return.columns.difference(drop)]
        else:
            dict_col = {v:k+1 for k,v in enumerate(df.columns)}
            columns_value = df.columns.difference(columns_nest).tolist()
            list_df_tmp = []
            for x in df.itertuples():
                df_tmp = pd.concat([x[dict_col[col]] for col in columns_nest], axis=1)
                for col in columns_value:
                    df_tmp[col] = x[dict_col[col]]
                list_df_tmp.append(df_tmp)
            df_return = pd.concat(list_df_tmp)
            return df_return[pd.Index(columns_value).append(df_return.columns.difference(columns_value))]
    else:
        column_series = df.columns[df.applymap(lambda x: isinstance(x, (pd.Series, np.ndarray, list))).sum() > 0].tolist()
        assert len(column_series) == 1, "Must exist one list of list Series"
        repeat_times = df[column_series[0]].map(len)
        columns_group = df.columns.difference(column_series)
        df_return = pd.DataFrame(df[columns_group].as_matrix().repeat(repeat_times, axis=0), columns=columns_group)
        df_series = pd.concat([*df[column_series[0]].map(lambda x: pd.DataFrame(x))], axis=0)
        df_return[column_series[0]] = df_series[0].tolist()
        if copy:
            return cp.deepcopy(df_return[df_return.columns.difference(drop)])
        else:
            return df_return[df_return.columns.difference(drop)]

def apply_window(df, func, partition=None, columns=None):
    """ apply window function in DataFrame

    Parameters
    ----------
    df: DataFrameGroupBy or DataFrame
    func: list of function
    partition: list of partition columns
    columns: list of columns which need to apply func
    """
    if isinstance(df, pd.core.groupby.DataFrameGroupBy):
        df_g = df
    else:
        df_g = df.groupby(partition)
    if columns is not None:
        if callable(func):
            if isinstance(columns, str):
                return df_g[columns].transform(func)
            elif isinstance(columns, (list, pd.core.indexes.base.Index)):
                df_return = df_g[columns].transform(func)
                df_return.columns = df_return.columns + "_" + func.__name__
                return df_return
        if isinstance(func, list):
            list_df = []
            df_return = pd.DataFrame()
            for v in func:
                if isinstance(columns, str):
                    return df_g[columns].transform(v)
                df_tmp = df_g[columns].transform(v)
                df_tmp.columns = df_tmp.columns + "_" + v.__name__
                list_df.append(df_tmp)
            return pd.concat(list_df, axis=1)
    if isinstance(func, dict):
        df_return = pd.DataFrame()
        for (k, v) in func.items():
            if isinstance(v, list):
                for each_fun in v:
                    df_return[k + '_' + each_fun.__name__] = df_g[k].transform(each_fun)
            else:
                df_return[k + '_' + v.__name__] = df_g[k].transform(v)
        if isinstance(df, pd.core.frame.DataFrame):
            df_return.index = df.index
        return df_return

def _series_to_dict(x, index_name='index'):
    """
    Change Pandas Series to Dict With Index

    Parameters
    ----------
    x : pandas Series
    index_name : return dict key of index name
    """
    x_dict = x.to_dict()
    x_dict[index_name] = x.name
    return x_dict

def to_dataframe(data, index_name='index'):
    """
    Change list of Pandas Serice to Pandas DataFrame

    Parameters
    ----------
    data : list of pandas Series
    index_name : return index DataFrame column name
    """
    p_series_to_dict = partial(_series_to_dict, index_name=index_name)
    return pd.DataFrame(list(map(p_series_to_dict, data)))



""" All combination rows from list of DataFrame """

import pandas as pd

def combination(dfs):
    """
    All combination rows from list of DataFrame

    Parameters
    ----------
    dfs : list of Pandas DataFrame

    Returns
    -------
    df_return : Pandas DataFrame
    """
    dfs_index_name = [df.index.name if df.index.name else 'index' for df in dfs]
    dfs = [df.reset_index().rename(columns={dfs_index_name[i]:dfs_index_name[i] + "_" + str(i)}) 
           if dfs_index_name[i] == 'index'
           else df.reset_index() 
           for i, df in enumerate(dfs)]
    join_key = list(pd.compat.product(*[df.index for df in  dfs ]))
    list_combine_df = []
    for i, df in enumerate(dfs):
        list_combine_df.append(dfs[i].iloc[[ x[i] for x in join_key],:].reset_index())
    df_return = pd.concat(list_combine_df , axis=1, ignore_index=False)
    del df_return['index']
    return df_return


import pandas as pd
import numpy as np
from sklearn import datasets
from tidyframe import nest, unnest, apply_window

iris = datasets.load_iris()
df = pd.DataFrame(iris['data'], columns=iris.feature_names)
df['target'] = iris.target
df['target2'] = list(map(lambda x: str(x +1),df.target))
field_apply = {k:v for (k, v) in (zip(df.columns[:2], [np.max, np.mean]))}
field_apply['sepal length (cm)'] = [np.min, np.max]
df_g = df.groupby('target')

def test_apply_window_basic():
    apply_window(df, np.mean, partition=['target'], columns=df.columns[1])

def test_apply_window_list_columns():
    apply_window(df, np.mean, partition=['target'], columns=df.columns[:3])

def test_apply_window_list_func():
    apply_window(df, field_apply, partition='target')

def test_apply_window_DataFrameGroupBy():
    apply_window(df.groupby('target'), field_apply).head()

def test_apply_window_all_list():
    apply_window(df_g, [np.mean, np.max], columns=['sepal length (cm)', 'sepal length (cm)'])

def test_apply_window_lsit_func_and_str_column():
    apply_window(df_g, [np.mean, np.max], columns='sepal length (cm)')






import math
import numpy as np
from tidyframe import Possibly

@Possibly()
def log_possibly(x):
    return math.log(x)

def test_Possibly_basic_success():
    assert np.isclose(log_possibly(10), math.log(10)), 'Must result is True'

def test_pPossibly_basic_fail():
    assert np.isnan(log_possibly(-10)), 'Must result is True'

def test_Possibly_change_otherwise():
    @Possibly(otherwise=-1)
    def log_possibly(x):
        return math.log(x)
    assert np.isclose(log_possibly(-10), -1), 'Must result is True'

def test_Possibly_classmethod_basic_success():
    assert np.isclose(Possibly.possibly(math.log)(10), math.log(10)), 'Must result is True'

def test_Possibly_classmethod_basic_fail():
    assert np.isnan(Possibly.possibly(math.log)(-1)), 'Must result is True'

def test_Possibly_classmethod_change_default():
    Possibly.otherwise_all = 1
    Possibly.quiet_all = False
    assert np.isclose(Possibly.possibly(math.log)(-1), 1), 'Must result is True'

def test_Possibly_print_exception():
    @Possibly(otherwise=-1, quiet=False)
    def log_possibly(x):
        return math.log(x)
    log_possibly(-10)



import math
import numpy as np
from tidyframe import Safely

@Safely()
def log_safely(x):
    return math.log(x)

def test_Safely_basic_success():
    result_log = log_safely(10)
    assert np.isclose(result_log['result'], math.log(10)), 'Must result be True'
    assert result_log['error'] is None, 'Must result be None'

def test_Safely_basic_faie():
    result_log2 = log_safely(-10)
    assert np.isnan(result_log2['result']), 'Must result is True'
    assert result_log2['error'] is not None, 'Must result is True'    

def test_Safely_classmethod_success():
    result_log3 = Safely.safely(math.log)(10)
    assert np.isclose(result_log3['result'], math.log(10)), 'Must result is True'
    assert result_log3['error'] is None, 'Must result is True'

def test_Safe_classmethod_fail():
    result_log4 = Safely.safely(math.log)(-1)
    assert np.isnan(result_log4['result']), 'Must result is True'
    assert result_log4['error'] is not None, 'Must result is True'

def test_Safely_classmethod_change_default():
    Safely.otherwise_all = -1
    Safely.quiet_all = False
    result_log5 = Safely.safely(math.log)(-1)
    assert np.isclose(result_log5['result'], -1), 'Must result is True'
    assert result_log5['error'] is not None, 'Must result is True'

def test_Safely_print_exception():
    @Safely(otherwise=-1, quiet=False)
    def log_safely(x):
        return math.log(x)
    log_safely(-10)



from tidyframe.tools.string import *

def test_replace_by_dict_uppdercase():
    map_uppercase_letters = { x:y for x,y  in zip(uppercase_letters, uppercase_fullwidth_letters)}
    replace_by_dict('/'.join(uppercase_letters), map_uppercase_letters)

def test_replace_by_dict_lowercase():
    map_lowercase_letters = { x:y for x,y  in zip(lowercase_letters, lowercase_fullwidth_letters)}
    replace_by_dict('/'.join(lowercase_letters), map_lowercase_letters)

def test_replace_by_dict_digits():
    map_digits = { x:y for x,y  in zip(digits, fullwidth_digits)}
    replace_by_dict('/'.join(digits), map_digits)



from .transform import nest, unnest, apply_window, to_dataframe
from .reshape import gather, spread
from .cature import Safely, Possibly
import tidyframe.tools as tools

__all__ = [
    "nest",
    "unnest",
    "apply_window",
    "to_dataframe",
    "gather",
    "spread",
    "Safely",
    "Possibly",
    "tools"
]



from tidyframe import tools
import pandas as pd
import numpy as np

df = pd.DataFrame(np.array(range(10)).reshape(2,5), columns=list('abcde'), index=['row_1', 'row_2'])

def test_select_basic():
    tools.select(df, columns=['b', 'd'])

def test_select_columns_minus():
    tools.select(df, columns_minus=['b', 'd'])

def test_select_deepcopy():
    tools.select(df, columns_minus=['b', 'd'], copy=True)

def test_select_columns_minus():
    tools.select(df, columns_minus=['b', 'd'])

def test_select_columns_between():
    tools.select(df, columns_between=['b', 'd'])


from .select import select, reorder_columns
from .separate import separate
from .combination import combination
from .string import replace_by_dict
from .dict import flatten_dict

__all__ = ['select' , 
           'reorder_columns', 
           'separate', 
           'combination', 
           'replace_by_dict', 
           'flatten_dict']



""" Deal with Function Exceptions """

import numpy as np

class Possibly(object):
    """Cature Exception if function exception occurred

    if function result occur exception then this decorator help you cature
    and return specified value(defaule: numpy.NaN)
    """
    otherwise_all = np.NaN
    quiet_all = True

    def __init__(self, otherwise=np.NaN, quiet=True):
        self.otherwise = otherwise
        self.quiet = quiet

    def __call__(self, func):
        def result_func(*args, **kargs):
            """ Catch all Function Exception """
            try:
                result = func(*args, **kargs)
            except Exception as e:
                result = self.otherwise
                if not self.quiet:
                    print(e)
            return result
        return result_func

    @classmethod
    def possibly(cls, func):
        """ class method for Catch all Function Exception """
        def result_func(*args, **kargs):
            try:
                result = func(*args, **kargs)
            except Exception as e:
                result = cls.otherwise_all
                if not cls.quiet_all:
                    print(e)
            return result
        return result_func

class Safely(object):
    """Cature Exception if function exception occurred

    This decorator will change function result to dict with two key.
    result: function result if no exception occurred, None if exception occurred
    error: None if no exception occurred, exception string if exception occurred
    """
    otherwise_all = np.NaN
    quiet_all = True

    def __init__(self, otherwise=np.NaN, quiet=True):
        self.otherwise = otherwise
        self.quiet = quiet

    def __call__(self, func):
        def result_func(*args, **kargs):
            """ Catch all Function Exception """
            result_dict = {}
            try:
                result_dict['result'] = func(*args, **kargs)
                result_dict['error'] = None
            except Exception as e:
                result_dict['result'] = self.otherwise
                result_dict['error'] = str(e)
                if not self.quiet:
                    print(e)
            return result_dict
        return result_func

    @classmethod
    def safely(cls, func):
        """ class method for Catch all Function Exception """
        def result_func(*args, **kargs):
            result_dict = {}
            try:
                result_dict['result'] = func(*args, **kargs)
                result_dict['error'] = None
            except Exception as e:
                result_dict['result'] = cls.otherwise_all
                result_dict['error'] = str(e)
                if not cls.quiet_all:
                    print(e)
            return result_dict
        return result_func



import pandas as pd
from tidyframe.tools import combination

df_a = pd.DataFrame({'a1':list('ABC'), 'a2':list('CDE')})
df_b = pd.DataFrame({'b1':list('01234'), 'b2':list('56789')})
df_c = pd.DataFrame({'c1':list('pq'), 'c2':list('rs')})
df_d = pd.DataFrame({'d1':list('abcd'), 'd2':list('efgh')})
df_d.index.name = 'index'

def test_combination_basic():
    combination([df_a, df_b, df_c])

def test_combination_basic_2():
    combination([df_a, df_b, df_d])



import pandas as pd 
import numpy as np
from tidyframe.tools import separate

df = pd.DataFrame({'full_string':['a b c d e z', 'f g h i']}, index=['row_1', 'row_2'])
series = df.full_string.str.split(' ')

def test_separate_basic():
    separate(series)

def test_separate_index():
    separate(series, index=[0, 4])

def test_separate_using_otherwise():
    separate(series, index=[0, 4], otherwise='otherwise')

def test_separate_change_column_name():
    separate(series, index=[0, 3], columns = ['zero', 'three'])

def test_separate_list_to_dataframe():
    separate([list('abc'), list('def')])



import pandas as pd
from sklearn import datasets
from tidyframe import gather

iris = datasets.load_iris()
df = pd.DataFrame(iris['data'], columns=iris.feature_names)
df['target'] = iris.target
df['target2'] = list(map(lambda x: 1 if x<1 else 0, df.target))
col_gather = df.columns[:4]
df_short = df.head()

def test_gather_basic():
    gather(df_short[col_gather].reset_index().head(8), col_gather ).head()

def test_gather_without_index():
    tmp = gather(df_short[col_gather])
    assert tmp.columns[0] == 'index'

def test_gather_assign_col():
    gather(df_short, col_gather )

def test_gather_str_key():
    gather(df_short, 'target')

def test_gather_with_index_name():
    df_short2 = df_short[col_gather]
    df_short2.index.name = 'index_with_name'
    assert gather(df_short2, col_gather ).columns[0] == df_short2.index.name




""" Convert Pandas DataFrame Between Wide Format and Long Format """

import pandas as pd

def gather(df, key_col=None, key='key', value='value', dropna=True):
    """Gather column to key-value pairs

    Parameters
    ----------
    df : DataFrame
    key : return DataFrame column name of key
    value : return DataFrame column name fo value
    dropna : boolean, default True
             Whether to drop rows in the resulting Frame/Series with no valid values

    Returns
    -------
    df_return : Pandas DataFrame
    """
    if key_col is None:
        key_col = df.columns
        index = []
        df = df.reset_index()
    elif isinstance(key_col, str):
        key_col = [key_col]
        df = df.reset_index()
        index = df.columns.difference(key_col).values.tolist()
    elif isinstance(key_col, (list, pd.core.indexes.base.Index)):
        index = df.columns.difference(key_col).values.tolist()
    if len(index) > 0:
        df_tmp = df.set_index(index)
        df_return = df_tmp[key_col].stack(dropna=dropna).reset_index()
        columns_return = df_return.columns.values.copy()
        columns_return[-1] = value
        columns_return[-2] = key
        df_return.columns = columns_return
        return df_return
    else:
        df_tmp = df
        raw_index_name = df.index.name
        if raw_index_name is None:
            raw_index_name = 'index'
        df_return = df_tmp[key_col].stack(dropna=dropna)
        df_return = df_return.to_frame().reset_index()
        df_return.columns = [raw_index_name, key, value]
        return df_return

def spread(df, row_index, key):
    """
    Spread key-value pair to multiple columns

    Parameters
    ----------
    df : long format Dataframe
    row_index : transform to wide format row index column
    key : key column which return DataFrame column name

    Returns
    -------
    df_return : Pandas DataFrame
    """
    list_index = []
    if isinstance(row_index, list):
        list_index.extend(row_index)
    else:
        list_index.append(row_index)
    if isinstance(key, list):
        list_index.extend(key)
        list_key = key
    else:
        list_index.append(key)
        list_key = [key]
    df_return = df.set_index(list_index)
    if len(list_key) == 1:
        return df_return.unstack(-1)
    else:
        index_start = len(list_index) - len(list_key)
        index_end = len(list_index) - 1
        return df_return.unstack(list(range(index_start, index_end + 1)))



import pandas as pd
from tidyframe.tools import flatten_dict

dict_1 = {'a':1, 'b': [1, 2], 'c':{'cc1':3, 'cc2':4}, 'd':{'d1':5, 'd2':{'dd1':6, 'dd2':7}}}
dict_2 = {'a':1, 'b': [1, 2], 'c':{'cc1':3}, 'd':{'d1':5, 'd2':{'dd1':6, 'dd2':7}}}
list_dict = [dict_1, dict_2]

def test_basic_flatten_dict():
    pd.DataFrame([flatten_dict(x) for x in list_dict])

def test_basic_flatten_dict_2():
    pd.DataFrame([flatten_dict(x, inner_name=True) for x in list_dict])



import pandas as pd
from tidyframe import to_dataframe

list_series = [pd.Series([1, 2], index=['i_1', 'i_2']),
               pd.Series([3, 4], index=['i_1', 'i_2'])]
list_series2 = [pd.Series([1, 2], index=['i_1', 'i_2']),
               pd.Series([3, 4], index=['i_3', 'i_4'])]

def test_to_dataframe_basic():
    assert to_dataframe(list_series).shape == (2 , 3)


def test_to_dataframe_index():
    assert to_dataframe(list_series2).shape == (2 , 5)

def test_to_dataframe_index_name():
    df = to_dataframe(list_series, index_name='index2')
    assert 'index2' in df.columns.tolist()



""" deal with dict object """

from copy import copy
from functools import reduce
from funcy import get_in

def flatten_dict(source_dict, name_delimiter='_', inner_name=False):
    """
    flatten nest dict

    Parameters
    ----------
    source_dict : nest dict
    name_delimiter : flatten name delimiter(non-use when inner_name is True)
    inner_name : False, use innermost name as retrun dict key or not 

    Returns
    -------
    flatten dict
    """
    assert isinstance(source_dict, dict), "import source_dict is not dict"
    json_name = {}
    for key in source_dict.keys():
        if isinstance(get_in(source_dict, [key]), dict):
            val = [True, [key]]
            json_name.update({key: val})
        else:
            val = [False, [key]]
            json_name.update({key: val})
    while True:
        key_inner = list(filter(lambda x: json_name.get(x)[0], json_name))
        if key_inner:
            for x in key_inner:
                dict_to_update_json_name = {}
                val = json_name.get(x)[1]
                for key in get_in(source_dict, val).keys():
                    val_in = copy(val)
                    val_in.append(key)
                    if isinstance(get_in(source_dict, val_in), dict):
                        dict_to_update = {reduce(lambda x, y: x + name_delimiter + y, val_in):[True, val_in]}
                    else:
                        dict_to_update = {reduce(lambda x, y: x + name_delimiter + y, val_in):[False, val_in]}
                    dict_to_update_json_name.update(dict_to_update)
                json_name.update(dict_to_update_json_name)
                json_name.pop(x)
        else:
            break
    if inner_name:
        return {json_name.get(x)[1][-1]:get_in(source_dict, json_name.get(x)[1]) for x in json_name.keys()}
    else:
        return {x:get_in(source_dict, json_name.get(x)[1]) for x in json_name.keys()}



import pandas as pd
from tidyframe.tools import flatten_dict, reorder_columns

dict_1 = {'a':1, 'b': [1, 2], 'c':{'cc1':3, 'cc2':4}, 'd':{'d1':5, 'd2':{'dd1':6, 'dd2':7}}}
dict_2 = {'a':1, 'b': [1, 2], 'c':{'cc1':3}, 'd':{'d1':5, 'd2':{'dd1':6, 'dd2':7}}}
list_dict = [dict_1, dict_2]
df = pd.DataFrame([flatten_dict(x) for x in list_dict])

def test_basic_reorder_columns():
    columns = ['d_d2_dd1', 'd_d2_dd2']
    reorder_columns(df, columns)

def test_basic_reorder_columns_pattern():
    reorder_columns(df, pattern='^d')



""" Easy Select Column Method from Pandas DataFrame """

import re
from copy import copy, deepcopy

def select(df, columns=None, columns_minus=None, columns_between=None, copy=False):
    """
    Select Pandas DataFrame by minus

    Parameters
    ----------
    df : Pandas DataFrame
    columns_minus : column which want to remove
    copy : whether return deep copy DataFrame

    Returns
    -------
    df_return : Pandas DataFrame
    """
    if columns is not None:
        df_return = df[columns]
    if columns_minus is not None:
        raw_col = {value:i for i, value in enumerate(df.columns)}
        for pop_key in columns_minus:
            raw_col.pop(pop_key)
        df_return = df[list(raw_col.keys())]
    if columns_between is not None:
        columns_location = {column:i for i, column in enumerate(df.columns) }
        assert columns_location[columns_between[0]] < columns_location[columns_between[1]], 'first column location must less than second column location'
        df_return = df.iloc[:,range(columns_location[columns_between[0]], columns_location[columns_between[1]] + 1)]
    if copy:
        return deepcopy(df_return)
    else:
        return df_return

def reorder_columns(df, columns=None, pattern=None):
    """
    reorder columns of pandas DataFrame

    Parameters
    ----------
    df : Pandas DataFrame
    columns : list which want to head column name(non-use if pattern is not None)
    pattern : regular expression pattern which let selected columns be at head columns

    Returns
    -------
    df_return : Pandas DataFrame
    """
    if pattern:
        reorder_columns = list(filter(lambda x: re.search(pattern, x), df.columns))
    else:
        reorder_columns = copy(list(columns))
    raw_columns = df.columns.copy()
    reorder_columns.extend(raw_columns.difference(reorder_columns).tolist())
    return df[reorder_columns]



""" string function  """

import copy

def replace_by_dict(source_string, mapping):
    """
    replace string by dictionary

    Parameters
    ----------
    source_string : string
    mapping : dict

    Returns
    -------
    return_string : string
    """
    return_string = copy.copy(source_string)
    for key, value in mapping.items():
        return_string = return_string.replace(key, value)
    return return_string

uppercase_letters = [chr(x) for x in range(65, 91)]
uppercase_fullwidth_letters = [chr(x) for x in range(65313, 65339)]
lowercase_letters = [chr(x) for x in range(97, 123)]
lowercase_fullwidth_letters = [chr(x) for x in range(65345, 65371)]
digits = [chr(x) for x in range(48, 58)]
fullwidth_digits = [chr(x) for x in range(65296, 65306)]
ord_symbol = [33, 34, 35, 36, 37,
              38, 39, 40, 41, 42,
              43, 44, 45, 46, 47,
              58, 59, 60, 61, 62,
              63, 91, 92, 93, 94,
              95, 96, 123, 124, 125,
              126, 40, 41, 65377, 65378,
              65379, 44, 46]
ord_fullwidth_symbol = []
ord_fullwidth_symbol.extend(range(65281, 65296))
ord_fullwidth_symbol.extend(range(65306, 65312))
ord_fullwidth_symbol.extend(range(65339, 65345))
ord_fullwidth_symbol.extend(range(65371, 65382))
symbol = [chr(x) for x in ord_symbol]
fullwidth_symbol = [chr(x) for x in ord_fullwidth_symbol]
ord_whitespace = [9, 10, 11, 12, 13,
                  32, 133, 160, 8192, 8193,
                  8194, 8195, 8196, 8197, 8198,
                  8199, 8200, 8201, 8202, 8232,
                  8233, 8239, 8287, 12288]
whitespace = [chr(x) for x in ord_whitespace]



from setuptools import setup, find_packages

LONG_DESCRIPTION = """
+ Clean pandas DataFrame to tidy DataFrame
+ Inspired by tidyr in R
+ Goal: Help you to create **tidy Pandas Dataframe**
+ Wapper window function from DataFrame
+ Implement R function *tidyr::gather* and *tidyr:spread* using python

"""

setup(
    name='tidyframe',
    version='0.0.1',
    author='Hsueh-Hung Cheng',
    author_email='jhengsh.email@gmail.com',
    url='https://github.com/Jhengsh/tidyframe',
    description='Clean pandas DataFrame to Tidy DataFrame',
    long_description=LONG_DESCRIPTION,
    keywords=['pandas', 'tidy'],
    packages=find_packages(),
    license='MIT',
    platforms='any',
    python_requires='!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    install_requires=["pandas"],
)



import pandas as pd
from sklearn import datasets
from tidyframe import nest, unnest

iris = datasets.load_iris()
df = pd.DataFrame(iris['data'], columns=iris.feature_names)
df['target'] = iris.target
df['target2'] = list(map(lambda x: str(x +1),df.target))
columns = ['target', 'target2']
df2 = nest(df, columns, key='data_nest', copy=True)

def test_unnest_basic():
    unnest(df2)[df.columns]

def test_unnest_list():
    df_string = pd.DataFrame()
    df_string['a'] = list('11223')
    df_string['b'] = list('22334')
    df_string['c'] = ['a', 'bb', 'ccc','dddd','eeeee']
    df_string['d'] = df_string['c'].map(lambda x: list(x))
    unnest(df_string)

def test_unnest_split():
    df_comma = pd.DataFrame({'a':list('12'), "b":['A,B,C', 'D,E,F,G']})
    df_comma['b_split'] = df_comma['b'].apply(lambda x: x.split(','))
    unnest(df_comma)

def test_deepcopy():
    unnest(df2, copy=True)

def test_unnest_list():
    df_string = pd.DataFrame()
    df_string['a'] = list('11223')
    df_string['b'] = list('22334')
    df_string['c'] = ['a', 'bb', 'ccc','dddd','eeeee']
    df_string['d'] = df_string['c'].map(lambda x: list(x))
    unnest(df_string, drop=['a'], copy=True)

def test_unnest_mulit_nest_dataframe():
    df2['data_nest2'] = df2['data_nest']
    unnest(df2)



import pandas as pd
from sklearn import datasets
from tidyframe import nest

iris = datasets.load_iris()
df = pd.DataFrame(iris['data'], columns=iris.feature_names)
df['target'] = iris.target
df['target2'] = list(map(lambda x: str(x +1),df.target))
columns = df.columns[:4].tolist()

def test_nest_columns():
    nest(df, columns, key='data_nest', copy=True)

def test_nest_columns():
    nest(df, columns_minus=df.columns[4:6].tolist(), key='data_nest', copy=True)

def test_nest_group_by():
    nest(df.groupby(['target', 'target2']), key='data_nest', copy=True)

def test_nest_by_index():
    nest(df, df.columns[:4], key='data_nest', copy=True)

def test_nest_by_index_minus():
    nest(df, columns_minus=df.columns[4:6], key='data_nest', copy=True).head()

def test_nest_column_between():
    nest(df, columns_between=['sepal length (cm)', 'petal width (cm)'], key='data_nest')



""" Separate string list to Pandas DataFrame """

import pandas as pd
import numpy as np

def _select_index(x, i, otherwise=np.NaN):
    """
    Select by index and Catch all Exception

    Parameters
    ----------
    x : array
    i : index
    otherwise : fill value if exist exception

    Returns
    -------
    x[i] if not exception happen else return otherwise
    """
    try:
        return x[i]
    except:
        return otherwise

def separate(series, index=None, columns=None, otherwise=np.NaN):
    """
    Separate string list to Pandas DataFrame

    Parameters
    ----------
    series : list of list or Series of list
    index : filter return index
    columns : return column name of DataFrame
    otherwise : numpy.NaN, fill value of not exist value

    Returns
    -------
    df_return : Pandas DataFrame with split each element of series to column
    """
    series = pd.Series(series)
    ncol = series.apply(len).max()
    if index is not None:
        assert max(index) < ncol, 'max of index MUST less than max of Series'
        if columns is not None:
            assert len(columns) == len(index), "length of columns MUST SAME as length of index"
        else:
            columns = pd.Series(['col'] * len(index)) + '_' + pd.Series(index).apply(str)
    else:
        index = list(range(ncol))
        columns = pd.Series(['col']*ncol) + '_' + pd.Series(range(ncol)).apply(str)
    return_df = pd.DataFrame()

    for i, name in zip(index, columns):
        return_df[name] = series.apply(lambda x: _select_index(x, i, otherwise))
    return return_df
